/*
 * Copyright (C) 2018 Purism SPC
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#ifndef __CHATTY_PURPLE_NOTIFY_H_INCLUDE__
#define __CHATTY_PURPLE_NOTIFY_H_INCLUDE__

#include <gtk/gtk.h>
#include <gtk/gtkwidget.h>
#include "notify.h"


PurpleNotifyUiOps *chatty_notify_get_ui_ops (void);

#endif
