/*
 * Copyright (C) 2018 Purism SPC
 * Copyright (C) 2019 Ruslan N. Marchenko
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#ifndef __XEPS_0313_H_INCLUDE__
#define __XEPS_0313_H_INCLUDE__

#define MAM_PREFS_DEF "mam-prefs-default"
#define MAM_DEF_DISABLE "disabled"
#define MAM_DEF_ROSTER  "roster"
#define MAM_DEF_ALWAYS  "always"
#define MAM_DEF_NEVER   "never"

void chatty_0313_init (void);
void chatty_0313_close (void);

#endif
