/* chatty-contact-private.h
 *
 * Copyright 2020 Purism SPC
 *
 * Author(s):
 *   Mohammed Sadiq <sadiq@sadiqpk.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include "chatty-contact.h"

void chatty_contact_clear_cache (ChattyContact *self);
