/*
 * Copyright (C) 2020 Purism SPC
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#pragma once

#include <glib.h>

G_BEGIN_DECLS

void chatty_0352_init (void);
void chatty_0352_close (void);

G_END_DECLS
