/*
 * Copyright (C) 2018 Purism SPC
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#ifndef __XEPS_0184_H_INCLUDE__
#define __XEPS_0184_H_INCLUDE__


void chatty_0184_init (void);
void chatty_0184_close (void);

#endif
