.topic_no_edit {
  border-style: solid;
  border-width: 1px;
  border-radius: 0px;
  transition: border-color 0.25s linear;
  transition: border-width 0.25s linear;
  background: @theme_base_color;
  border-color: alpha(@theme_fg_color, 0.25);
}

.topic_edit {
  border-style: solid;
  border-width: 2px;
  border-radius: 3px;
  transition: border-color 0.25s linear;
  transition: border-width 0.25s linear;
  background: @theme_base_color;
  border-color: @theme_selected_bg_color;
}

.button_send_green:not(:disabled) {
  background: #6cba3d;
  border-color: #6cba3d;
  color: white;
  outline-color: rgba(255, 255, 255, 0.3);
  -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.543529);
  box-shadow: inset 0 1px rgba(255, 255, 255, 0.4);
}

.box_message {
  background-color: @theme_base_color;
  border-bottom: 1px solid @borders;
}

.list_no_select:selected {
  background-color: @theme_base_color;
}

.list_select:selected {
  background-color: #c0bfbc;
}


/* MESSAGE LIST AND BUBBLES */

.message_bubble {
  font-size: 16px;
  color: black;
  border-radius: 9px;
  box-shadow: 0 1px 2px rgba(0, 0, 0, 0.4);
  padding-top: 3px;
  padding-bottom: 3px;
  padding-left: 10px;
  padding-right: 10px;
}

.bubble_white {
  background-color: white;
}

.bubble_blue {
  background-color: #e4f0fc;
}

.bubble_green {
  background-color: #e2fae7;
}

.bubble_purple {
  font-size: 14px;
  color: white;
  background-color: #862d86;
}

.label_disclaim {
  font-size: 16px;
  font-weight: bold;
  padding-top: 120px;
  padding-bottom: 20px;
}

.unread-badge {
  background-color: @theme_selected_bg_color;
  color: white;
  font-weight: bold;
  font-size: 0.8em;
  border-radius: 9999px;
  min-width: 0.7em;
  padding: 2px 6px;
}

.muted-badge {
  background-color: #bfbfbf;
  color: white;
}

row:selected .unread-badge {
  background-color: white;
  color: @theme_selected_bg_color;
}

.bold-label {
  font-weight: bold;
}

@define-color protocol_tag_bg #c12c87;
@define-color protocol_tag_fg #ffffff;

.protocol-tag {
  background-image: none;
  background-color: @protocol_tag_bg;
  color: @protocol_tag_fg;

  border-radius: 4px;
  border-width: 0;

  margin: 2px;
  padding: 4px;
}

.protocol-tag:hover {
    background-color: shade(@protocol_tag_bg, 1.10);
    color: @protocol_tag_fg;
}

.protocol-tag.button,
.protocol-tag.button:hover,
.protocol-tag.button:active,
.protocol-tag.button:active:hover {
  background-color: transparent;
  background-image: none;
  box-shadow: none;
  border-image: none;
  border-width: 0;
}

.protocol-tag.button:hover {
    color: shade(@protocol_tag_bg, 2.10);
}

popover.no-padding {
  padding: 0px;
}

button.delete-button {
  padding: 0px;
  min-height: 27px;
  min-width: 27px;
}

.no-margin,
.no-margin > scrolledwindow > viewport > clamp {
  margin: 0;
}
