((nil . ((indent-tabs-mode . nil)))
 (c-mode . ((c-file-style . "GNU")
            (c-basic-offset . 2)
            (c-file-offsets
             (brace-list-intro . +))))
 (c-ts-mode . ((c-ts-mode-style . gnu)
               (c-ts-mode-indent-offset . 2)))
 )
